package com.supervision.rasa.service.impl;

import cn.hutool.core.collection.CollectionUtil;
import cn.hutool.core.collection.ListUtil;
import cn.hutool.core.io.FileUtil;
import cn.hutool.core.util.StrUtil;
import com.supervision.exception.BusinessException;
import com.supervision.model.RasaModelInfo;
import com.supervision.rasa.config.ThreadPoolExecutorConfig;
import com.supervision.rasa.constant.RasaConstant;
import com.supervision.rasa.pojo.vo.RasaCmdArgumentVo;
import com.supervision.rasa.service.RasaCmdService;
import com.supervision.rasa.util.PortUtil;
import com.supervision.service.RasaModeService;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.core.io.ClassPathResource;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.io.*;
import java.util.*;
import java.util.concurrent.*;
import java.util.function.Predicate;

@Service
@Slf4j
@RequiredArgsConstructor
public class RasaCmdServiceImpl implements RasaCmdService {


    @Value("${rasa.models-path}")
    private String modelsPath;

    @Value("${rasa.endpoints}")
    private String endpoints;

    @Value("${rasa.config}")
    private String config;

    @Value("${rasa.data-path}")
    private String dataPath;


    @Value("${rasa.shell-env:/bin/bash}")
    private String shellEnv;

    private final RasaModeService rasaModeService;


    private final ConcurrentHashMap<String,String> shellPathCache = new ConcurrentHashMap<>();


    @Override
    @Transactional
    public String trainExec(RasaCmdArgumentVo argument) throws ExecutionException, InterruptedException, TimeoutException {

        argument.setFixedModelNameIfAbsent();

        // /rasa/v3_jiazhuangxian/domain.yml  domain的路径，应该是从zip文件中加压出来的文件的路径后面拼上/domain.yml
        String domain = replaceDuplicateSeparator(String.join(File.separator,dataPath,argument.getModelId(),"domain.yml"));

        // /rasa/v3_jiazhuangxian/   yml文件的路径，应该是从zip文件中加压出来的文件的路径,在配置文件中配置
        String localDataPath = replaceDuplicateSeparator(String.join(File.separator,dataPath,argument.getModelId()));

        //  /rasa/models 生成出来的模型的存放路径,也写在配置文件里面
        String localModelsPath = replaceDuplicateSeparator(String.join(File.separator,modelsPath,argument.getModelId()));

        List<String> cmds = ListUtil.toList(shellEnv, getShellPath(RasaConstant.TRAIN_SHELL),config,localDataPath,domain,localModelsPath);

        cmds.add(argument.getFixedModelName());

        log.info("trainExec cmd : {}",StrUtil.join(" ",cmds));

        List<String> outMessage = execCmd(cmds, s -> StrUtil.isNotBlank(s) && s.contains(RasaConstant.TRAN_SUCCESS_MESSAGE), 300);

        String outMessageString = String.join("\r\n", outMessage);
        //保存 模型信息
        RasaModelInfo rasaModelInfo = new RasaModelInfo();
        rasaModelInfo.setModelId(argument.getModelId());
        rasaModelInfo.setTranStatus(trainIsSuccess(outMessage)?1:0);
        rasaModelInfo.setServerStatus(-1);
        cmds.set(1,null);
        rasaModelInfo.setTrainCmd(cmds);
        rasaModelInfo.setTrainLog(outMessageString);
        rasaModeService.saveOrUpdateByModelId(rasaModelInfo);

        return outMessageString;
    }



    @Override
    public String runExec(RasaCmdArgumentVo argument) throws ExecutionException, InterruptedException, TimeoutException {


        // 1. 查找可用端口
        int port = PortUtil.findUnusedPort(5050, 100000,rasaModeService.listActivePort());
        log.info("runExec findUnusedPort is : {}",port);

        // 2. 运行模型
        // 2.1 -m 参数对应的值 /rasa/models/aaa1111.tar.gz 指run的模型的路径，/rasa/models应该来自于配置文件，和训练时的--out是同一配置项；
        // aaa1111.tar.gz这个，前面的文件名应该是--fixed-model-name指定的，.tar.gz是文件后缀，代码拼接

        String fixedModePath;
        String modeParentPath = replaceDuplicateSeparator(String.join(File.separator, modelsPath, argument.getModelId()));
        if (StrUtil.isEmpty(argument.getFixedModelName())){
              fixedModePath = listLastFilePath(modeParentPath, f -> f.getName().matches("-?\\d+(\\.\\d+)?.tar.gz"));
        }else {
            String fixedModelName = argument.getFixedModelName()+".tar.gz";
            fixedModePath = String.join(modeParentPath,fixedModelName);
        }

        List<String> cmds = ListUtil.toList(shellEnv, getShellPath(RasaConstant.RUN_SHELL),fixedModePath,endpoints,String.valueOf(port));

        log.info("runExec cmd : {}",StrUtil.join(" ",cmds));

        List<String> outMessageList = execCmd(cmds, s -> StrUtil.isNotBlank(s) && s.contains(RasaConstant.RUN_SUCCESS_MESSAGE), 300);

        String outMessageString = String.join("\r\n", outMessageList);

        // 3. 尝试停止原先的程序
        boolean runIsSuccess = runIsSuccess(outMessageList);
        RasaModelInfo dbRasaModelInfo = rasaModeService.queryByModelId(argument.getModelId());
        if (null != dbRasaModelInfo.getPort() && runIsSuccess){
            log.info("runExec server is up and dbRasaModelInfo is: {}, start kill old server ....",dbRasaModelInfo.getPort());
            // 3.1 如果原本的程序已经运行且本次成功启动，杀掉原本的进程
            execCmd(ListUtil.toList(shellEnv, getShellPath(RasaConstant.KILL_SHELL), String.valueOf(dbRasaModelInfo.getPort())), s -> false, 20);
        }
        if (null != dbRasaModelInfo.getPort() && !runIsSuccess){
            // 3.2 如果原本的程序已经运行，但本次程序未正常启动，不对端口号进行变动
            log.info("runExec server start fail and dbRasaModelInfo is: {}, user old server ....",dbRasaModelInfo.getPort());
            port = dbRasaModelInfo.getPort();
        }
        // 4. 更新模型信息
        RasaModelInfo rasaModelInfo = new RasaModelInfo();
        rasaModelInfo.setModelId(argument.getModelId());
        rasaModelInfo.setPort(port);
        rasaModelInfo.setServerStatus( runIsSuccess ? 1 : 0 );
        rasaModelInfo.setRunCmd(ListUtil.toList(shellEnv, null,fixedModePath,endpoints));
        rasaModelInfo.setRunLog(outMessageString);
        rasaModeService.saveOrUpdateByModelId(rasaModelInfo);

        return outMessageString;
    }



    public List<String> execCmd(List<String> cmds, Predicate<String> endPredicate, long timeOut) throws InterruptedException, ExecutionException, TimeoutException {
        ProcessBuilder processBuilder = new ProcessBuilder(cmds);
        processBuilder.redirectErrorStream(true);

        ThreadPoolExecutor instance = ThreadPoolExecutorConfig.getInstance();
        Future<List<String>> future = instance.submit(() -> {
            Process process = processBuilder.start();
            InputStream inputStream = process.getInputStream();
            Reader reader = new InputStreamReader(inputStream, "UTF-8");
            BufferedReader bufferedReader = new BufferedReader(reader);
            List<String> outString = new ArrayList<>();
            String resultLines = bufferedReader.readLine();
            while( resultLines != null) {
                resultLines = bufferedReader.readLine();
                log.info("resultLines:{}",resultLines);
                outString.add(resultLines);
                if (endPredicate.test(resultLines)){
                    break;
                }
            }
            bufferedReader.close();
            return outString;
        });

        return future.get(timeOut, TimeUnit.SECONDS);
    }




    private  boolean trainIsSuccess(List<String> messageList){

        return containKey(messageList,RasaConstant.TRAN_SUCCESS_MESSAGE);
    }


    private  boolean runIsSuccess(List<String> messageList){

        return containKey(messageList,RasaConstant.RUN_SUCCESS_MESSAGE);
    }

    private boolean containKey(List<String> messageList,String keyWord){

        if (CollectionUtil.isEmpty(messageList)){
            return false;
        }
        if (StrUtil.isEmpty(keyWord)){
            return false;
        }
        return messageList.stream().anyMatch(s->StrUtil.isNotEmpty(s) && s.contains(keyWord));
    }

    private String replaceDuplicateSeparator(String path){

        if (StrUtil.isEmpty(path)){
            return path;
        }

       return path.replace(File.separator + File.separator, File.separator);
    }



    private String listLastFilePath(String path, FileFilter filter){
        File file = listLastFile(path, filter);
        if (null == file){
            return null;
        }
        return file.getPath();
    }
    private File listLastFile(String path,FileFilter filter){
        File file = new File(path);
        File[] files = file.listFiles(filter);
        if (null == files){
            return null;
        }

        return Arrays.stream(files).max(Comparator.comparing(File::getName)).orElse(null);
    }


    public String getShellPath(String shell){
        if (StrUtil.isEmpty(shell)){
            throw new BusinessException("shell 不能为空...");
        }

        // 1. 先从缓存中获取shell脚本路径
        String shellPath = "shellScript/" + shell;
        String cachePath = shellPathCache.get(shell);
        if (StrUtil.isNotEmpty(cachePath)){
            if (new File(cachePath).exists()){
                return cachePath;
            }else {
                shellPathCache.remove(shell);
            }
        }

        // 2. 如果缓存中不存在shell脚本路径，从resources目录下加载
        ClassPathResource shellScript = new ClassPathResource(shellPath);

        try {
            // 2.1 加载文件，写入临时目录中
            InputStream inputStream = shellScript.getInputStream();
            File tempFile = File.createTempFile(shell, ".sh");
            tempFile.deleteOnExit();
            FileUtil.writeFromStream(inputStream, tempFile);
            // 2.2 给文件赋可执行权限
            ProcessBuilder chmod = new ProcessBuilder("chmod", "+x", shellPath);
            Process chmodProcess = chmod.start();
            chmodProcess.waitFor();

            // 2.3 把文件路径放到缓存信息中
            shellPathCache.put(shell,tempFile.getAbsolutePath());

            return tempFile.getAbsolutePath();
        } catch (IOException | InterruptedException e) {
            throw new RuntimeException(e);
        }
    }


}
